# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

"""State writer implementation."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from airbyte_protocol.models.airbyte_protocol import AirbyteStateMessage


class StateWriterBase(abc.ABC):
    """A class to write state artifacts."""

    @abc.abstractmethod
    def write_state(
        self,
        state_message: AirbyteStateMessage,
    ) -> None:
        """Save or 'write' a state artifact."""
        ...


class StdOutStateWriter(StateWriterBase):
    """A state writer that writes state artifacts to stdout.

    This is required when we are functioning as a "Destination" in the Airbyte protocol, and
    an orchestrator is responsible for saving those state artifacts.
    """

    def write_state(
        self,
        state_message: AirbyteStateMessage,
    ) -> None:
        """Save or 'write' a state artifact."""
        print(state_message.model_dump_json())



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Experimental features which may change.

> **NOTE:**
> The below "experimental" features are now "stable" and can be accessed directly from the
`airbyte.get_source()` method.

The experimental `get_source` implementation allows you to run sources
using Docker containers. This feature is still in development and may
change in the future.

To use this feature, import `get_source` from this module and use it in place of the `get_source`
function from the `airbyte` module.

Instead of this:

```python
from airbyte import ab

source = ab.get_source(...)
```

Use this:

```python
from airbyte.experimental import get_source

source = get_source(...)
```

Experimental features may change without notice between minor versions of PyAirbyte. Although rare,
they may also be entirely removed or refactored in future versions of PyAirbyte. Experimental
features may also be less stable than other features, and may not be as well-tested.

You can help improve this product by reporting issues and providing feedback for improvements in our
[GitHub issue tracker](https://github.com/airbytehq/pyairbyte/issues).
"""

from __future__ import annotations

from airbyte.sources.util import get_source


__all__ = [
    "get_source",
]



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""SQL processors."""

from __future__ import annotations

from airbyte._processors.sql import snowflakecortex
from airbyte._processors.sql.snowflakecortex import (
    SnowflakeCortexSqlProcessor,
    SnowflakeCortexTypeConverter,
)


__all__ = [
    # Classes
    "SnowflakeCortexSqlProcessor",
    "SnowflakeCortexTypeConverter",
    # modules
    "snowflakecortex",
]



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""A Postgres implementation of the PyAirbyte cache.

## Usage Example

```python
from airbyte as ab
from airbyte.caches import PostgresCache

cache = PostgresCache(
    host="myhost",
    port=5432,
    username="myusername",
    password=ab.get_secret("POSTGRES_PASSWORD"),
    database="mydatabase",
)
```
"""

from __future__ import annotations

from pydantic import PrivateAttr

from airbyte._processors.sql.postgres import PostgresConfig, PostgresSqlProcessor
from airbyte.caches.base import CacheBase


class PostgresCache(PostgresConfig, CacheBase):
    """Configuration for the Postgres cache.

    Also inherits config from the JsonlWriter, which is responsible for writing files to disk.
    """

    _sql_processor_class = PrivateAttr(default=PostgresSqlProcessor)



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Catalog provider implementation.

A catalog provider wraps a configured catalog and configured streams. This class is responsible for
providing information about the catalog and streams. A catalog provider can also be updated with new
streams as they are discovered, providing a thin layer of abstraction over the configured catalog.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, final

from airbyte import exceptions as exc


if TYPE_CHECKING:
    from airbyte_protocol.models import (
        ConfiguredAirbyteCatalog,
        ConfiguredAirbyteStream,
    )


class CatalogProvider:
    """A catalog provider wraps a configured catalog and configured streams.

    This class is responsible for providing information about the catalog and streams.

    Note:
    - The catalog provider is not responsible for managing the catalog or streams but it may
      be updated with new streams as they are discovered.
    """

    def __init__(
        self,
        configured_catalog: ConfiguredAirbyteCatalog,
    ) -> None:
        """Initialize the catalog manager with a catalog object reference.

        Since the catalog is passed by reference, the catalog manager may be updated with new
        streams as they are discovered.
        """
        self._catalog: ConfiguredAirbyteCatalog = configured_catalog

    @property
    def configured_catalog(self) -> ConfiguredAirbyteCatalog:
        return self._catalog

    @property
    def stream_names(self) -> list[str]:
        return list({stream.stream.name for stream in self.configured_catalog.streams})

    def get_configured_stream_info(
        self,
        stream_name: str,
    ) -> ConfiguredAirbyteStream:
        """Return the column definitions for the given stream."""
        if not self.configured_catalog:
            raise exc.PyAirbyteInternalError(
                message="Cannot get stream JSON schema without a catalog.",
            )

        matching_streams: list[ConfiguredAirbyteStream] = [
            stream
            for stream in self.configured_catalog.streams
            if stream.stream.name == stream_name
        ]
        if not matching_streams:
            raise exc.AirbyteStreamNotFoundError(
                stream_name=stream_name,
                context={
                    "available_streams": [
                        stream.stream.name for stream in self.configured_catalog.streams
                    ],
                },
            )

        if len(matching_streams) > 1:
            raise exc.PyAirbyteInternalError(
                message="Multiple streams found with same name.",
                context={
                    "stream_name": stream_name,
                },
            )

        return matching_streams[0]

    @final
    def get_stream_json_schema(
        self,
        stream_name: str,
    ) -> dict[str, Any]:
        """Return the column definitions for the given stream."""
        return self.get_configured_stream_info(stream_name).stream.json_schema

    def get_stream_properties(
        self,
        stream_name: str,
    ) -> dict[str, dict]:
        """Return the names of the top-level properties for the given stream."""
        return self.get_stream_json_schema(stream_name)["properties"]



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""A Snowflake vector store implementation of the SQL processor."""

from __future__ import annotations

from textwrap import dedent, indent
from typing import TYPE_CHECKING

import sqlalchemy
from overrides import overrides
from sqlalchemy import text

from airbyte._processors.sql.snowflake import (
    SnowflakeConfig,
    SnowflakeSqlProcessor,
    SnowflakeTypeConverter,
)


if TYPE_CHECKING:
    from pathlib import Path

    from sqlalchemy.engine import Connection

    from airbyte._future_cdk.catalog_providers import CatalogProvider
    from airbyte._future_cdk.state_writers import StateWriterBase
    from airbyte._processors.file.base import FileWriterBase


class SnowflakeCortexTypeConverter(SnowflakeTypeConverter):
    """A class to convert array type into vector."""

    def __init__(
        self,
        conversion_map: dict | None = None,
        *,
        vector_length: int,
    ) -> None:
        self.vector_length = vector_length
        super().__init__(conversion_map)

    @overrides
    def to_sql_type(
        self,
        json_schema_property_def: dict[str, str | dict | list],
    ) -> sqlalchemy.types.TypeEngine:
        """Convert a value to a SQL type."""
        sql_type = super().to_sql_type(json_schema_property_def)
        if isinstance(sql_type, sqlalchemy.types.ARRAY):
            # SQLAlchemy doesn't yet support the `VECTOR` data type.
            # We may want to remove this or update once this resolves:
            # https://github.com/snowflakedb/snowflake-sqlalchemy/issues/499
            return f"VECTOR(FLOAT, {self.vector_length})"

        return sql_type


class SnowflakeCortexSqlProcessor(SnowflakeSqlProcessor):
    """A Snowflake implementation for use with Cortex functions."""

    supports_merge_insert = True
    type_converter_class = SnowflakeCortexTypeConverter

    def __init__(
        self,
        *,
        vector_length: int,
        catalog_provider: CatalogProvider,
        state_writer: StateWriterBase | None = None,
        sql_config: SnowflakeConfig,
        file_writer: FileWriterBase | None = None,
        temp_dir: Path | None = None,
        temp_file_cleanup: bool = True,
    ) -> None:
        """Custom initialization: Initialize type_converter with vector_length."""
        self._vector_length = vector_length

        # call base class to do necessary initialization
        super().__init__(
            catalog_provider=catalog_provider,
            state_writer=state_writer,
            sql_config=sql_config,
            file_writer=file_writer,
            temp_dir=temp_dir,
            temp_file_cleanup=temp_file_cleanup,
        )

    def _get_column_list_from_table(
        self,
        table_name: str,
    ) -> list[str]:
        """Get column names for passed stream.

        This is overridden due to lack of SQLAlchemy compatibility for the
        `VECTOR` data type.
        """
        conn: Connection = self.sql_config.get_vendor_client()
        cursor = conn.cursor()
        cursor.execute(f"DESCRIBE TABLE {table_name};")
        results = cursor.fetchall()
        column_names = [row[0].lower() for row in results]
        cursor.close()
        conn.close()
        return column_names

    @overrides
    def _write_files_to_new_table(
        self,
        files: list[Path],
        stream_name: str,
        batch_id: str,
    ) -> str:
        """Write files to a new table."""
        temp_table_name = self._create_table_for_loading(
            stream_name=stream_name,
            batch_id=batch_id,
        )
        internal_sf_stage_name = f"@%{temp_table_name}"

        def path_str(path: Path) -> str:
            return str(path.absolute()).replace("\\", "\\\\")

        put_files_statements = "\n".join(
            [f"PUT 'file://{path_str(file_path)}' {internal_sf_stage_name};" for file_path in files]
        )
        self._execute_sql(put_files_statements)
        columns_list = [
            self._quote_identifier(c)
            for c in list(self._get_sql_column_definitions(stream_name).keys())
        ]
        files_list = ", ".join([f"'{f.name}'" for f in files])
        columns_list_str: str = indent("\n, ".join(columns_list), " " * 12)

        # following two lines are different from SnowflakeSqlProcessor
        vector_suffix = f"::Vector(Float, {self._vector_length})"
        variant_cols_str: str = ("\n" + " " * 21 + ", ").join(
            [
                f"$1:{self.normalizer.normalize(col)}{vector_suffix if 'embedding' in col else ''}"
                for col in columns_list
            ]
        )

        copy_statement = dedent(
            f"""
            COPY INTO {temp_table_name}
            (
                {columns_list_str}
            )
            FROM (
                SELECT {variant_cols_str}
                FROM {internal_sf_stage_name}
            )
            FILES = ( {files_list} )
            FILE_FORMAT = ( TYPE = JSON )
            ;
            """
        )
        self._execute_sql(copy_statement)
        return temp_table_name

    @overrides
    def _add_missing_columns_to_table(
        self,
        stream_name: str,
        table_name: str,
    ) -> None:
        """Use Snowflake Python connector to add new columns to the table"""
        columns = self._get_sql_column_definitions(stream_name)
        existing_columns = self._get_column_list_from_table(table_name)
        for column_name, column_type in columns.items():
            if column_name not in existing_columns:
                self._add_new_column_to_table(table_name, column_name, column_type)
            self._invalidate_table_cache(table_name)
        pass

    def _add_new_column_to_table(
        self,
        table_name: str,
        column_name: str,
        column_type: sqlalchemy.types.TypeEngine,
    ) -> None:
        conn: Connection = self.sql_config.get_vendor_client()
        cursor = conn.cursor()
        cursor.execute(
            text(
                f"ALTER TABLE {self._fully_qualified(table_name)} "
                f"ADD COLUMN {column_name} {column_type}"
            ),
        )
        cursor.close()
        conn.close()



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""SQL Cache implementation."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, final

import pandas as pd
import pyarrow as pa
import pyarrow.dataset as ds
from pydantic import Field, PrivateAttr

from airbyte_protocol.models import ConfiguredAirbyteCatalog

from airbyte._future_cdk.catalog_providers import CatalogProvider
from airbyte._future_cdk.sql_processor import (
    SqlConfig,
    SqlProcessorBase,
)
from airbyte._future_cdk.state_writers import StdOutStateWriter
from airbyte.caches._catalog_backend import CatalogBackendBase, SqlCatalogBackend
from airbyte.caches._state_backend import SqlStateBackend
from airbyte.constants import DEFAULT_ARROW_MAX_CHUNK_SIZE
from airbyte.datasets._sql import CachedDataset


if TYPE_CHECKING:
    from collections.abc import Iterator

    from airbyte._future_cdk.sql_processor import SqlProcessorBase
    from airbyte._future_cdk.state_providers import StateProviderBase
    from airbyte._future_cdk.state_writers import StateWriterBase
    from airbyte.caches._state_backend_base import StateBackendBase
    from airbyte.datasets._base import DatasetBase


class CacheBase(SqlConfig):
    """Base configuration for a cache.

    Caches inherit from the matching `SqlConfig` class, which provides the SQL config settings
    and basic connectivity to the SQL database.

    The cache is responsible for managing the state of the data synced to the cache, including the
    stream catalog and stream state. The cache also provides the mechanism to read and write data
    to the SQL backend specified in the `SqlConfig` class.
    """

    cache_dir: Path = Field(default=Path(".cache"))
    """The directory to store the cache in."""

    cleanup: bool = True
    """Whether to clean up the cache after use."""

    _deployed_api_root: Optional[str] = PrivateAttr(default=None)
    _deployed_workspace_id: Optional[str] = PrivateAttr(default=None)
    _deployed_destination_id: Optional[str] = PrivateAttr(default=None)

    _sql_processor_class: type[SqlProcessorBase] = PrivateAttr()
    _read_processor: SqlProcessorBase = PrivateAttr()

    _catalog_backend: CatalogBackendBase = PrivateAttr()
    _state_backend: StateBackendBase = PrivateAttr()

    def __init__(self, **data: Any) -> None:  # noqa: ANN401
        """Initialize the cache and backends."""
        super().__init__(**data)

        # Create a temporary processor to do the work of ensuring the schema exists
        temp_processor = self._sql_processor_class(
            sql_config=self,
            catalog_provider=CatalogProvider(ConfiguredAirbyteCatalog(streams=[])),
            state_writer=StdOutStateWriter(),
            temp_dir=self.cache_dir,
            temp_file_cleanup=self.cleanup,
        )
        temp_processor._ensure_schema_exists()  # noqa: SLF001  # Accessing non-public member

        # Initialize the catalog and state backends
        self._catalog_backend = SqlCatalogBackend(
            engine=self.get_sql_engine(),
            table_prefix=self.table_prefix or "",
        )
        self._state_backend = SqlStateBackend(
            engine=self.get_sql_engine(),
            table_prefix=self.table_prefix or "",
        )

        # Now we can create the SQL read processor
        self._read_processor = self._sql_processor_class(
            sql_config=self,
            catalog_provider=self._catalog_backend.get_full_catalog_provider(),
            state_writer=StdOutStateWriter(),  # Shouldn't be needed for the read-only processor
            temp_dir=self.cache_dir,
            temp_file_cleanup=self.cleanup,
        )

    @final
    @property
    def processor(self) -> SqlProcessorBase:
        """Return the SQL processor instance."""
        return self._read_processor

    def get_record_processor(
        self,
        source_name: str,
        catalog_provider: CatalogProvider,
    ) -> SqlProcessorBase:
        """Return a record processor for the specified source name and catalog.

        We first register the source and its catalog with the catalog manager. Then we create a new
        SQL processor instance with (only) the given input catalog.

        For the state writer, we use a state writer which stores state in an internal SQL table.
        """
        # First register the source and catalog into durable storage. This is necessary to ensure
        # that we can later retrieve the catalog information.
        self.register_source(
            source_name=source_name,
            incoming_source_catalog=catalog_provider.configured_catalog,
            stream_names=set(catalog_provider.stream_names),
        )

        # Next create a new SQL processor instance with the given catalog - and a state writer
        # that writes state to the internal SQL table and associates with the given source name.
        return self._sql_processor_class(
            sql_config=self,
            catalog_provider=catalog_provider,
            state_writer=self.get_state_writer(source_name=source_name),
            temp_dir=self.cache_dir,
            temp_file_cleanup=self.cleanup,
        )

    # Read methods:

    def get_records(
        self,
        stream_name: str,
    ) -> CachedDataset:
        """Uses SQLAlchemy to select all rows from the table."""
        return CachedDataset(self, stream_name)

    def get_pandas_dataframe(
        self,
        stream_name: str,
    ) -> pd.DataFrame:
        """Return a Pandas data frame with the stream's data."""
        table_name = self._read_processor.get_sql_table_name(stream_name)
        engine = self.get_sql_engine()
        return pd.read_sql_table(table_name, engine, schema=self.schema_name)

    def get_arrow_dataset(
        self,
        stream_name: str,
        *,
        max_chunk_size: int = DEFAULT_ARROW_MAX_CHUNK_SIZE,
    ) -> ds.Dataset:
        """Return an Arrow Dataset with the stream's data."""
        table_name = self._read_processor.get_sql_table_name(stream_name)
        engine = self.get_sql_engine()

        # Read the table in chunks to handle large tables which does not fits in memory
        pandas_chunks = pd.read_sql_table(
            table_name=table_name,
            con=engine,
            schema=self.schema_name,
            chunksize=max_chunk_size,
        )

        arrow_batches_list = []
        arrow_schema = None

        for pandas_chunk in pandas_chunks:
            if arrow_schema is None:
                # Initialize the schema with the first chunk
                arrow_schema = pa.Schema.from_pandas(pandas_chunk)

            # Convert each pandas chunk to an Arrow Table
            arrow_table = pa.RecordBatch.from_pandas(pandas_chunk, schema=arrow_schema)
            arrow_batches_list.append(arrow_table)

        return ds.dataset(arrow_batches_list)

    @final
    @property
    def streams(self) -> dict[str, CachedDataset]:
        """Return a temporary table name."""
        result = {}
        stream_names = set(self._catalog_backend.stream_names)

        for stream_name in stream_names:
            result[stream_name] = CachedDataset(self, stream_name)

        return result

    def get_state_provider(
        self,
        source_name: str,
        *,
        refresh: bool = True,
    ) -> StateProviderBase:
        """Return a state provider for the specified source name."""
        return self._state_backend.get_state_provider(
            source_name=source_name,
            table_prefix=self.table_prefix or "",
            refresh=refresh,
        )

    def get_state_writer(
        self,
        source_name: str,
    ) -> StateWriterBase:
        """Return a state writer for the specified source name."""
        return self._state_backend.get_state_writer(source_name=source_name)

    def register_source(
        self,
        source_name: str,
        incoming_source_catalog: ConfiguredAirbyteCatalog,
        stream_names: set[str],
    ) -> None:
        """Register the source name and catalog."""
        self._catalog_backend.register_source(
            source_name=source_name,
            incoming_source_catalog=incoming_source_catalog,
            incoming_stream_names=stream_names,
        )

    def __getitem__(self, stream: str) -> DatasetBase:
        """Return a dataset by stream name."""
        return self.streams[stream]

    def __contains__(self, stream: str) -> bool:
        """Return whether a stream is in the cache."""
        return stream in (self._catalog_backend.stream_names)

    def __iter__(  # type: ignore [override]  # Overriding Pydantic model method
        self,
    ) -> Iterator[tuple[str, Any]]:
        """Iterate over the streams in the cache."""
        return ((name, dataset) for name, dataset in self.streams.items())



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""A BigQuery implementation of the cache.

## Usage Example

```python
import airbyte as ab
from airbyte.caches import BigQueryCache

cache = BigQueryCache(
    project_name="myproject",
    dataset_name="mydataset",
    credentials_path="path/to/credentials.json",
)
```
"""

from __future__ import annotations

from pydantic import PrivateAttr

from airbyte._processors.sql.bigquery import BigQueryConfig, BigQuerySqlProcessor
from airbyte.caches.base import (
    CacheBase,
)
from airbyte.constants import DEFAULT_ARROW_MAX_CHUNK_SIZE


class BigQueryCache(BigQueryConfig, CacheBase):
    """The BigQuery cache implementation."""

    _sql_processor_class: type[BigQuerySqlProcessor] = PrivateAttr(default=BigQuerySqlProcessor)

    def get_arrow_dataset(
        self,
        stream_name: str,
        *,
        max_chunk_size: int = DEFAULT_ARROW_MAX_CHUNK_SIZE,
    ) -> None:
        """Raises NotImplementedError; BigQuery doesn't support `pd.read_sql_table`.
        https://github.com/airbytehq/PyAirbyte/issues/165
        """
        raise NotImplementedError(
            "BigQuery doesn't currently support to_arrow"
            "Please consider using a different cache implementation for these functionalities."
        )



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

"""State backend implementation."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, cast

from pytz import utc
from sqlalchemy import Column, DateTime, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from airbyte_protocol.models import (
    AirbyteStateMessage,
    AirbyteStateType,
)

from airbyte._future_cdk.state_providers import StaticInputState
from airbyte._future_cdk.state_writers import StateWriterBase
from airbyte.caches._state_backend_base import (
    StateBackendBase,
)
from airbyte.exceptions import PyAirbyteInternalError


if TYPE_CHECKING:
    from sqlalchemy.engine import Engine

    from airbyte._future_cdk.state_providers import StateProviderBase


STATE_TABLE_NAME = "_airbyte_state"

GLOBAL_STATE_STREAM_NAME = "_GLOBAL"
LEGACY_STATE_STREAM_NAME = "_LEGACY"
GLOBAL_STATE_STREAM_NAMES = [GLOBAL_STATE_STREAM_NAME, LEGACY_STATE_STREAM_NAME]


SqlAlchemyModel = declarative_base()
"""A base class to use for SQLAlchemy ORM models."""


class StreamState(SqlAlchemyModel):  # type: ignore[valid-type,misc]
    """A SQLAlchemy ORM model to store state metadata."""

    __tablename__ = STATE_TABLE_NAME

    source_name = Column(String)
    """The source name."""

    stream_name = Column(String)
    """The stream name."""

    table_name = Column(String, primary_key=True)
    """The table name holding records for the stream."""

    state_json = Column(String)
    """The JSON string representation of the state message."""

    last_updated = Column(
        DateTime(timezone=True), onupdate=datetime.now(utc), default=datetime.now(utc)
    )
    """The last time the state was updated."""


class SqlStateWriter(StateWriterBase):
    """State writer for SQL backends."""

    def __init__(self, source_name: str, backend: SqlStateBackend) -> None:
        self._state_backend: SqlStateBackend = backend
        self.source_name = source_name

    def write_state(
        self,
        state_message: AirbyteStateMessage,
    ) -> None:
        if state_message.type == AirbyteStateType.GLOBAL:
            stream_name = GLOBAL_STATE_STREAM_NAME
        if state_message.type == AirbyteStateType.LEGACY:
            stream_name = LEGACY_STATE_STREAM_NAME
        elif state_message.type == AirbyteStateType.STREAM:
            stream_name = state_message.stream.stream_descriptor.name
        else:
            raise PyAirbyteInternalError(
                message="Invalid state message type.",
                context={"state_message": state_message},
            )

        self._state_backend._ensure_internal_tables()  # noqa: SLF001  # Non-public member access
        table_prefix = self._state_backend._table_prefix  # noqa: SLF001
        engine = self._state_backend._engine  # noqa: SLF001

        with Session(engine) as session:
            session.query(StreamState).filter(
                StreamState.table_name == table_prefix + stream_name
            ).delete()

            # This prevents "duplicate key" errors but (in theory) should not be necessary.
            session.commit()
            session.add(
                StreamState(
                    source_name=self.source_name,
                    stream_name=stream_name,
                    table_name=table_prefix + stream_name,
                    state_json=state_message.model_dump_json(),
                )
            )
            session.commit()


class SqlStateBackend(StateBackendBase):
    """
    A class to manage the stream catalog of data synced to a cache:

    - What streams exist and to what tables they map
    - The JSON schema for each stream
    """

    def __init__(
        self,
        engine: Engine,
        table_prefix: str = "",
    ) -> None:
        """Initialize the state manager with a static catalog state."""
        self._engine: Engine = engine
        self._table_prefix = table_prefix
        super().__init__()

    def _ensure_internal_tables(self) -> None:
        """Ensure the internal tables exist in the SQL database."""
        engine = self._engine
        SqlAlchemyModel.metadata.create_all(engine)

    def get_state_provider(
        self,
        source_name: str,
        table_prefix: str = "",
        streams_filter: list[str] | None = None,
        *,
        refresh: bool = True,
    ) -> StateProviderBase:
        """Return the state provider.

        Subclasses may add additional keyword arguments to this method as needed.
        """
        _ = refresh  # Always refresh the state
        self._ensure_internal_tables()

        engine = self._engine
        with Session(engine) as session:
            query = session.query(StreamState).filter(
                StreamState.source_name == source_name
                and (
                    StreamState.table_name.startswith(table_prefix)
                    or StreamState.stream_name.in_(GLOBAL_STATE_STREAM_NAMES)
                )
            )
            if streams_filter:
                query = query.filter(
                    StreamState.stream_name.in_([*streams_filter, *GLOBAL_STATE_STREAM_NAMES])
                )
            states: list[StreamState] = cast(list[StreamState], query.all())
            # Only return the states if the table name matches what the current cache
            # would generate. Otherwise consider it part of a different cache.
            states = [
                state for state in states if state.table_name == table_prefix + state.stream_name
            ]

        return StaticInputState(
            from_state_messages=[
                AirbyteStateMessage.model_validate_json(state.state_json) for state in states
            ]
        )

    def get_state_writer(
        self,
        source_name: str,
    ) -> StateWriterBase:
        return SqlStateWriter(source_name=source_name, backend=self)



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""PyAirbyte classes and methods for interacting with the Airbyte Cloud API.

You can use this module to interact with Airbyte Cloud, OSS, and Enterprise.

## Examples

### Basic Usage Example:

```python
import airbyte as ab
from airbyte import cloud

# Initialize an Airbyte Cloud workspace object
workspace = cloud.CloudWorkspace(
    workspace_id="123",
    api_key=ab.get_secret("AIRBYTE_CLOUD_API_KEY"),
)

# Run a sync job on Airbyte Cloud
connection = workspace.get_connection(connection_id="456")
sync_result = connection.run_sync()
print(sync_result.get_job_status())
```

### Example Read From Cloud Destination:

If your destination is supported, you can read records directly from the
`SyncResult` object. Currently this is supported in Snowflake and BigQuery only.


```python
# Assuming we've already created a `connection` object...

# Get the latest job result and print the stream names
sync_result = connection.get_sync_result()
print(sync_result.stream_names)

# Get a dataset from the sync result
dataset: CachedDataset = sync_result.get_dataset("users")

# Get a SQLAlchemy table to use in SQL queries...
users_table = dataset.to_sql_table()
print(f"Table name: {users_table.name}")

# Or iterate over the dataset directly
for record in dataset:
    print(record)
```

ℹ️ **Experimental Features**

You can use the `airbyte.cloud.experimental` module to access experimental features.
These additional features are subject to change and may not be available in all environments.
"""  # noqa: RUF002  # Allow emoji

from __future__ import annotations

from airbyte.cloud import connections, constants, sync_results, workspaces
from airbyte.cloud.connections import CloudConnection
from airbyte.cloud.constants import JobStatusEnum
from airbyte.cloud.sync_results import SyncResult
from airbyte.cloud.workspaces import CloudWorkspace


__all__ = [
    # Submodules
    "workspaces",
    "connections",
    "constants",
    "sync_results",
    # Classes
    "CloudWorkspace",
    "CloudConnection",
    "SyncResult",
    # Enums
    "JobStatusEnum",
]



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Utility functions for working with sources."""

from __future__ import annotations

import shutil
import sys
import tempfile
import warnings
from json import JSONDecodeError
from pathlib import Path
from typing import Any, cast

import requests
import yaml

from airbyte import exceptions as exc
from airbyte._executor import DockerExecutor, PathExecutor, VenvExecutor
from airbyte._util.telemetry import EventState, log_install_state
from airbyte.sources.base import Source
from airbyte.sources.declarative import DeclarativeExecutor
from airbyte.sources.registry import ConnectorMetadata, get_connector_metadata


def get_connector(
    name: str,
    config: dict[str, Any] | None = None,
    *,
    version: str | None = None,
    pip_url: str | None = None,
    local_executable: Path | str | None = None,
    install_if_missing: bool = True,
) -> Source:
    """Deprecated. Use get_source instead."""
    warnings.warn(
        "The `get_connector()` function is deprecated and will be removed in a future version."
        "Please use `get_source()` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_source(
        name=name,
        config=config,
        version=version,
        pip_url=pip_url,
        local_executable=local_executable,
        install_if_missing=install_if_missing,
    )


def get_source(  # noqa: PLR0912, PLR0913, PLR0915 # Too complex
    name: str,
    config: dict[str, Any] | None = None,
    *,
    streams: str | list[str] | None = None,
    version: str | None = None,
    pip_url: str | None = None,
    local_executable: Path | str | None = None,
    docker_image: bool | str = False,
    use_host_network: bool = False,
    source_manifest: bool | dict | Path | str = False,
    install_if_missing: bool = True,
    install_root: Path | None = None,
) -> Source:
    """Get a connector by name and version.

    Args:
        name: connector name
        config: connector config - if not provided, you need to set it later via the set_config
            method.
        streams: list of stream names to select for reading. If set to "*", all streams will be
            selected. If not provided, you can set it later via the `select_streams()` or
            `select_all_streams()` method.
        version: connector version - if not provided, the currently installed version will be used.
            If no version is installed, the latest available version will be used. The version can
            also be set to "latest" to force the use of the latest available version.
        pip_url: connector pip URL - if not provided, the pip url will be inferred from the
            connector name.
        local_executable: If set, the connector will be assumed to already be installed and will be
            executed using this path or executable name. Otherwise, the connector will be installed
            automatically in a virtual environment.
        docker_image: If set, the connector will be executed using Docker. You can specify `True`
            to use the default image for the connector, or you can specify a custom image name.
            If `version` is specified and your image name does not already contain a tag
            (e.g. `my-image:latest`), the version will be appended as a tag (e.g. `my-image:0.1.0`).
        use_host_network: If set, along with docker_image, the connector will be executed using
            the host network. This is useful for connectors that need to access resources on
            the host machine, such as a local database. This parameter is ignored when
            `docker_image` is not set.
        source_manifest: If set, the connector will be executed based on a declarative YAML
            source definition. This input can be `True` to attempt to auto-download a YAML spec,
            `dict` to accept a Python dictionary as the manifest, `Path` to pull a manifest from
            the local file system, or `str` to pull the definition from a web URL.
        install_if_missing: Whether to install the connector if it is not available locally. This
            parameter is ignored when `local_executable` or `source_manifest` are set.
        install_root: (Optional.) The root directory where the virtual environment will be
            created. If not provided, the current working directory will be used.
    """
    if (
        sum(
            [
                bool(local_executable),
                bool(docker_image),
                bool(pip_url),
                bool(source_manifest),
            ]
        )
        > 1
    ):
        raise exc.PyAirbyteInputError(
            message=(
                "You can only specify one of the settings: 'local_executable', 'docker_image', "
                "'pip_url', or 'source_manifest'."
            ),
            context={
                "local_executable": local_executable,
                "docker_image": docker_image,
                "pip_url": pip_url,
                "source_manifest": source_manifest,
            },
        )

    if local_executable:
        if version:
            raise exc.PyAirbyteInputError(
                message="Param 'version' is not supported when 'local_executable' is set."
            )

        if isinstance(local_executable, str):
            if "/" in local_executable or "\\" in local_executable:
                # Assume this is a path
                local_executable = Path(local_executable).absolute()
            else:
                which_executable: str | None = None
                which_executable = shutil.which(local_executable)
                if not which_executable and sys.platform == "win32":
                    # Try with the .exe extension
                    local_executable = f"{local_executable}.exe"
                    which_executable = shutil.which(local_executable)

                if which_executable is None:
                    raise exc.AirbyteConnectorExecutableNotFoundError(
                        connector_name=name,
                        context={
                            "executable": local_executable,
                            "working_directory": Path.cwd().absolute(),
                        },
                    ) from FileNotFoundError(local_executable)
                local_executable = Path(which_executable).absolute()

        print(f"Using local `{name}` executable: {local_executable!s}")
        return Source(
            name=name,
            config=config,
            streams=streams,
            executor=PathExecutor(
                name=name,
                path=local_executable,
            ),
        )

    if docker_image:
        if docker_image is True:
            # Use the default image name for the connector
            docker_image = f"airbyte/{name}"

        if version is not None and ":" in docker_image:
            raise exc.PyAirbyteInputError(
                message="The 'version' parameter is not supported when a tag is already set in the "
                "'docker_image' parameter.",
                context={
                    "docker_image": docker_image,
                    "version": version,
                },
            )

        if ":" not in docker_image:
            docker_image = f"{docker_image}:{version or 'latest'}"

        temp_dir = tempfile.gettempdir()

        docker_cmd = [
            "docker",
            "run",
            "--rm",
            "-i",
            "--volume",
            f"{temp_dir}:{temp_dir}",
        ]

        if use_host_network is True:
            docker_cmd.extend(["--network", "host"])

        docker_cmd.extend([docker_image])

        return Source(
            name=name,
            config=config,
            streams=streams,
            executor=DockerExecutor(
                name=name,
                executable=docker_cmd,
            ),
        )

    if source_manifest:
        if source_manifest is True:
            # Auto-set the manifest to a valid http address URL string
            source_manifest = (
                "https://raw.githubusercontent.com/airbytehq/airbyte/master/airbyte-integrations"
                f"/connectors/{name}/{name.replace('-', '_')}/manifest.yaml"
            )
        if isinstance(source_manifest, str):
            print("Installing connector from YAML manifest:", source_manifest)
            # Download the manifest file
            response = requests.get(url=source_manifest)
            response.raise_for_status()  # Raise an exception if the download failed

            if "class_name:" in response.text:
                raise exc.AirbyteConnectorInstallationError(
                    message=(
                        "The provided manifest requires additional code files (`class_name` key "
                        "detected). This feature is not compatible with the declarative YAML "
                        "executor. To use this executor, please try again with the Python "
                        "executor."
                    ),
                    connector_name=name,
                    context={
                        "manifest_url": source_manifest,
                    },
                )

            try:
                source_manifest = cast(dict, yaml.safe_load(response.text))
            except JSONDecodeError as ex:
                raise exc.AirbyteConnectorInstallationError(
                    connector_name=name,
                    context={
                        "manifest_url": source_manifest,
                    },
                ) from ex

        if isinstance(source_manifest, Path):
            source_manifest = cast(dict, yaml.safe_load(source_manifest.read_text()))

        # Source manifest is a dict at this point
        return Source(
            name=name,
            config=config,
            streams=streams,
            executor=DeclarativeExecutor(
                manifest=source_manifest,
            ),
        )
    # else: we are installing a connector in a virtual environment:

    metadata: ConnectorMetadata | None = None
    try:
        metadata = get_connector_metadata(name)
    except exc.AirbyteConnectorNotRegisteredError as ex:
        if not pip_url:
            log_install_state(name, state=EventState.FAILED, exception=ex)
            # We don't have a pip url or registry entry, so we can't install the connector
            raise

    try:
        executor = VenvExecutor(
            name=name,
            metadata=metadata,
            target_version=version,
            pip_url=pip_url,
            install_root=install_root,
        )
        if install_if_missing:
            executor.ensure_installation()

        return Source(
            name=name,
            config=config,
            streams=streams,
            executor=executor,
        )
    except Exception as e:
        log_install_state(name, state=EventState.FAILED, exception=e)
        raise


__all__ = [
    "get_source",
]



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

"""State provider implementation."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from airbyte_protocol.models import (
    AirbyteStateType,
)
from airbyte_protocol.models.airbyte_protocol import AirbyteStreamState

from airbyte import exceptions as exc


if TYPE_CHECKING:
    from airbyte_protocol.models import (
        AirbyteStateMessage,
        AirbyteStreamState,
    )


class StateProviderBase(abc.ABC):  # noqa: B024
    """A class to provide state artifacts."""

    def __init__(self) -> None:
        """Initialize the state manager with a static catalog state.

        This constructor may be overridden by subclasses to initialize the state artifacts.
        """
        self._state_message_artifacts: list[AirbyteStateMessage] | None = None

    @property
    def stream_state_artifacts(
        self,
    ) -> list[AirbyteStreamState]:
        """Return all state artifacts.

        This is just a type guard around the private variable `_stream_state_artifacts` and the
        cast to `AirbyteStreamState` objects.
        """
        if self._state_message_artifacts is None:
            raise exc.PyAirbyteInternalError(message="No state artifacts were declared.")

        return [
            state_msg.stream
            for state_msg in self._state_message_artifacts
            if state_msg.type == AirbyteStateType.STREAM
        ]

    @property
    def state_message_artifacts(
        self,
    ) -> list[AirbyteStreamState]:
        """Return all state artifacts.

        This is just a type guard around the private variable `_state_message_artifacts`.
        """
        if self._state_message_artifacts is None:
            raise exc.PyAirbyteInternalError(message="No state artifacts were declared.")

        return self._state_message_artifacts

    @property
    def known_stream_names(
        self,
    ) -> set[str]:
        """Return the unique set of all stream names with stored state."""
        return {state.stream_descriptor.name for state in self.stream_state_artifacts}

    def to_state_input_file_text(self) -> str:
        """Return the state artifacts as a JSON string.

        This is used when sending the state artifacts to the destination.
        """
        return (
            "["
            + "\n, ".join(
                [
                    state_artifact.model_dump_json()
                    for state_artifact in (self._state_message_artifacts or [])
                ]
            )
            + "]"
        )


class StaticInputState(StateProviderBase):
    """A state manager that uses a static catalog state as input."""

    def __init__(
        self,
        from_state_messages: list[AirbyteStateMessage] | None = None,
    ) -> None:
        """Initialize the state manager with a static catalog state."""
        self._state_message_artifacts: list[AirbyteStateMessage] | None = from_state_messages



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""A Snowflake implementation of the PyAirbyte cache.

## Usage Example

```python
from airbyte as ab
from airbyte.caches import SnowflakeCache

cache = SnowflakeCache(
    account="myaccount",
    username="myusername",
    password=ab.get_secret("SNOWFLAKE_PASSWORD"),
    warehouse="mywarehouse",
    database="mydatabase",
    role="myrole",
    schema_name="myschema",
)
```
"""

from __future__ import annotations

from pydantic import PrivateAttr

from airbyte._future_cdk.sql_processor import RecordDedupeMode
from airbyte._processors.sql.snowflake import SnowflakeConfig, SnowflakeSqlProcessor
from airbyte.caches.base import CacheBase


class SnowflakeCache(SnowflakeConfig, CacheBase):
    """Configuration for the Snowflake cache."""

    dedupe_mode: RecordDedupeMode = RecordDedupeMode.APPEND

    _sql_processor_class = PrivateAttr(default=SnowflakeSqlProcessor)



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""A DuckDB implementation of the PyAirbyte cache.

## Usage Example

```python
from airbyte as ab
from airbyte.caches import DuckDBCache

cache = DuckDBCache(
    db_path="/path/to/my/duckdb-file",
    schema_name="myschema",
)
```
"""

from __future__ import annotations

import warnings

from duckdb_engine import DuckDBEngineWarning
from pydantic import PrivateAttr

from airbyte._processors.sql.duckdb import DuckDBConfig, DuckDBSqlProcessor
from airbyte.caches.base import CacheBase


# Suppress warnings from DuckDB about reflection on indices.
# https://github.com/Mause/duckdb_engine/issues/905
warnings.filterwarnings(
    "ignore",
    message="duckdb-engine doesn't yet support reflection on indices",
    category=DuckDBEngineWarning,
)


# @dataclass
class DuckDBCache(DuckDBConfig, CacheBase):
    """A DuckDB cache."""

    _sql_processor_class: type[DuckDBSqlProcessor] = PrivateAttr(default=DuckDBSqlProcessor)



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Support for declarative yaml source testing."""

from __future__ import annotations

import json
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pydantic

from airbyte_cdk.entrypoint import AirbyteEntrypoint
from airbyte_cdk.sources.declarative.manifest_declarative_source import ManifestDeclarativeSource

from airbyte._executor import Executor
from airbyte.exceptions import PyAirbyteInternalError
from airbyte.sources.base import Source


if TYPE_CHECKING:
    from collections.abc import Iterator


def _suppress_cdk_pydantic_deprecation_warnings() -> None:
    """Suppress deprecation warnings from Pydantic in the CDK.

    CDK has deprecated uses of `json()` and `parse_obj()`, and we don't want users
    to see these warnings.
    """
    warnings.filterwarnings(
        "ignore",
        category=pydantic.warnings.PydanticDeprecatedSince20,
    )


class DeclarativeExecutor(Executor):
    """An executor for declarative sources."""

    def __init__(
        self,
        manifest: str | dict | Path,
    ) -> None:
        """Initialize a declarative executor.

        - If `manifest` is a path, it will be read as a json file.
        - If `manifest` is a string, it will be parsed as an HTTP path.
        - If `manifest` is a dict, it will be used as is.
        """
        _suppress_cdk_pydantic_deprecation_warnings()
        self._manifest_dict: dict
        if isinstance(manifest, Path):
            self._manifest_dict = cast(dict, json.loads(manifest.read_text()))

        elif isinstance(manifest, str):
            # TODO: Implement HTTP path parsing
            raise NotImplementedError("HTTP path parsing is not yet implemented.")

        elif isinstance(manifest, dict):
            self._manifest_dict = manifest

        if not isinstance(self._manifest_dict, dict):
            raise PyAirbyteInternalError(message="Manifest must be a dict.")

        self.declarative_source = ManifestDeclarativeSource(source_config=self._manifest_dict)
        self.reported_version: str | None = None  # TODO: Consider adding version detection

    def execute(self, args: list[str]) -> Iterator[str]:
        """Execute the declarative source."""
        source_entrypoint = AirbyteEntrypoint(self.declarative_source)
        parsed_args = source_entrypoint.parse_args(args)
        yield from source_entrypoint.run(parsed_args)

    def ensure_installation(self, *, auto_fix: bool = True) -> None:
        """No-op. The declarative source is included with PyAirbyte."""
        _ = auto_fix
        pass

    def install(self) -> None:
        """No-op. The declarative source is included with PyAirbyte."""
        pass

    def uninstall(self) -> None:
        """No-op. The declarative source is included with PyAirbyte."""
        pass


class DeclarativeSource(Source):
    """A declarative source using Airbyte's Yaml low-code/no-code framework."""

    def __init__(
        self,
        manifest: str | dict | Path,
    ) -> None:
        """Initialize a declarative source.

        Sample usages:
        ```python
        manifest_path = "path/to/manifest.yaml"

        source_a = DeclarativeSource(manifest=Path(manifest_path))
        source_b = DeclarativeSource(manifest=Path(manifest_path).read_text())
        source_c = DeclarativeSource(manifest=yaml.load(Path(manifest_path).read_text()))
        ```

        Args:
            manifest: The manifest for the declarative source. This can be a path to a yaml file, a
            yaml string, or a dict.
        """
        _suppress_cdk_pydantic_deprecation_warnings()

        # TODO: Conform manifest to a dict or str (TBD)
        self.manifest = manifest

        # Initialize the source using the base class implementation
        super().__init__(
            name="Declarative",  # TODO: Get name from manifest
            config={  # TODO: Put 'real' config here
                "manifest": manifest,
            },
            executor=DeclarativeExecutor(manifest),
        )



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Abstract base class for Processors, including SQL processors.

Processors accept Airbyte messages as input from STDIN or from another input stream.
"""

from __future__ import annotations

import abc
import io
import sys
from collections import defaultdict
from typing import TYPE_CHECKING, cast, final

from airbyte_protocol.models import (
    AirbyteMessage,
    AirbyteRecordMessage,
    AirbyteStateMessage,
    AirbyteStateType,
    AirbyteStreamState,
    Type,
)

from airbyte import exceptions as exc
from airbyte._future_cdk.state_writers import StdOutStateWriter
from airbyte.strategies import WriteStrategy


if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from airbyte._batch_handles import BatchHandle
    from airbyte._future_cdk.catalog_providers import CatalogProvider
    from airbyte._future_cdk.state_writers import StateWriterBase


class AirbyteMessageParsingError(Exception):
    """Raised when an Airbyte message is invalid or cannot be parsed."""


class RecordProcessorBase(abc.ABC):
    """Abstract base class for classes which can process Airbyte messages from a source.

    This class is responsible for all aspects of handling Airbyte protocol.

    The class should be passed a catalog manager and stream manager class to handle the
    catalog and state aspects of the protocol.
    """

    def __init__(
        self,
        *,
        catalog_provider: CatalogProvider,
        state_writer: StateWriterBase | None = None,
    ) -> None:
        """Initialize the processor.

        If a state writer is not provided, the processor will use the default (STDOUT) state writer.
        """
        self._catalog_provider: CatalogProvider | None = catalog_provider
        self._state_writer: StateWriterBase | None = state_writer or StdOutStateWriter()

        self._pending_state_messages: dict[str, list[AirbyteStateMessage]] = defaultdict(list, {})
        self._finalized_state_messages: dict[
            str,
            list[AirbyteStateMessage],
        ] = defaultdict(list, {})

        self._setup()

    @property
    def expected_streams(self) -> set[str]:
        """Return the expected stream names."""
        return set(self.catalog_provider.stream_names)

    @property
    def catalog_provider(
        self,
    ) -> CatalogProvider:
        """Return the catalog manager.

        Subclasses should set this property to a valid catalog manager instance if one
        is not explicitly passed to the constructor.

        Raises:
            PyAirbyteInternalError: If the catalog manager is not set.
        """
        if not self._catalog_provider:
            raise exc.PyAirbyteInternalError(
                message="Catalog manager should exist but does not.",
            )

        return self._catalog_provider

    @property
    def state_writer(
        self,
    ) -> StateWriterBase:
        """Return the state writer instance.

        Subclasses should set this property to a valid state manager instance if one
        is not explicitly passed to the constructor.

        Raises:
            PyAirbyteInternalError: If the state manager is not set.
        """
        if not self._state_writer:
            raise exc.PyAirbyteInternalError(
                message="State manager should exist but does not.",
            )

        return self._state_writer

    @final
    def process_stdin(
        self,
        *,
        write_strategy: WriteStrategy = WriteStrategy.AUTO,
    ) -> None:
        """Process the input stream from stdin.

        Return a list of summaries for testing.
        """
        input_stream = io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8")
        self.process_input_stream(
            input_stream,
            write_strategy=write_strategy,
        )

    @final
    def _airbyte_messages_from_buffer(
        self,
        buffer: io.TextIOBase,
    ) -> Iterator[AirbyteMessage]:
        """Yield messages from a buffer."""
        yield from (AirbyteMessage.model_validate_json(line) for line in buffer)

    @final
    def process_input_stream(
        self,
        input_stream: io.TextIOBase,
        *,
        write_strategy: WriteStrategy = WriteStrategy.AUTO,
    ) -> None:
        """Parse the input stream and process data in batches.

        Return a list of summaries for testing.
        """
        messages = self._airbyte_messages_from_buffer(input_stream)
        self.process_airbyte_messages(
            messages,
            write_strategy=write_strategy,
        )

    @abc.abstractmethod
    def process_record_message(
        self,
        record_msg: AirbyteRecordMessage,
        stream_schema: dict,
    ) -> None:
        """Write a record.

        This method is called for each record message.

        In most cases, the SQL processor will not perform any action, but will pass this along to to
        the file processor.
        """

    @final
    def process_airbyte_messages(
        self,
        messages: Iterable[AirbyteMessage],
        *,
        write_strategy: WriteStrategy,
    ) -> None:
        """Process a stream of Airbyte messages."""
        if not isinstance(write_strategy, WriteStrategy):
            raise exc.AirbyteInternalError(
                message="Invalid `write_strategy` argument. Expected instance of WriteStrategy.",
                context={"write_strategy": write_strategy},
            )

        stream_schemas: dict[str, dict] = {}

        # Process messages, writing to batches as we go
        for message in messages:
            if message.type is Type.RECORD:
                record_msg = cast(AirbyteRecordMessage, message.record)
                stream_name = record_msg.stream

                if stream_name not in stream_schemas:
                    stream_schemas[stream_name] = self.catalog_provider.get_stream_json_schema(
                        stream_name=stream_name
                    )

                self.process_record_message(
                    record_msg,
                    stream_schema=stream_schemas[stream_name],
                )

            elif message.type is Type.STATE:
                state_msg = cast(AirbyteStateMessage, message.state)
                if state_msg.type in {AirbyteStateType.GLOBAL, AirbyteStateType.LEGACY}:
                    self._pending_state_messages[f"_{state_msg.type}"].append(state_msg)
                else:
                    stream_state = cast(AirbyteStreamState, state_msg.stream)
                    stream_name = stream_state.stream_descriptor.name
                    self._pending_state_messages[stream_name].append(state_msg)

            else:
                # Ignore unexpected or unhandled message types:
                # Type.LOG, Type.TRACE, Type.CONTROL, etc.
                pass

        # We've finished processing input data.
        # Finalize all received records and state messages:
        self.write_all_stream_data(
            write_strategy=write_strategy,
        )

        self.cleanup_all()

    def write_all_stream_data(self, write_strategy: WriteStrategy) -> None:
        """Finalize any pending writes."""
        for stream_name in self.catalog_provider.stream_names:
            self.write_stream_data(
                stream_name,
                write_strategy=write_strategy,
            )

    @abc.abstractmethod
    def write_stream_data(
        self,
        stream_name: str,
        write_strategy: WriteStrategy,
    ) -> list[BatchHandle]:
        """Write pending stream data to the cache."""
        ...

    def _finalize_state_messages(
        self,
        state_messages: list[AirbyteStateMessage],
    ) -> None:
        """Handle state messages by passing them to the catalog manager."""
        if state_messages:
            self.state_writer.write_state(
                state_message=state_messages[-1],
            )

    def _setup(self) -> None:  # noqa: B027  # Intentionally empty, not abstract
        """Create the database.

        By default this is a no-op but subclasses can override this method to prepare
        any necessary resources.
        """
        pass

    def cleanup_all(self) -> None:  # noqa: B027  # Intentionally empty, not abstract
        """Clean up all resources.

        The default implementation is a no-op.
        """
        pass
